"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Session = void 0;
const fs = require("fs");
const chokidar = require("chokidar");
const path = require("path");
const getSessionConfiguration_1 = require("./getSessionConfiguration");
const write_1 = require("../util/write");
const parseCSS_1 = require("./parseCSS");
const extractPluginResult_1 = require("./extractPluginResult");
const generateScript_1 = require("./generateScript");
const waitForInitialScanCompletion_1 = require("./waitForInitialScanCompletion");
const minifyScripts_1 = require("../minifier/minifyScripts");
const createExposedPromise_1 = require("../util/createExposedPromise");
const minifyScriptsForCSS_1 = require("../minifier/minifyScriptsForCSS");
const deleteFile_1 = require("../util/deleteFile");
const writeFilep_1 = require("../util/writeFilep");
const { copyFile } = fs.promises;
class Session {
    constructor(parameters = {}) {
        this.configuration = getSessionConfiguration_1.getSessionConfiguration(parameters);
        this.processedFiles = new Set();
        this.initialTask = null;
        this.tasks = new Set();
    }
    get helperPath() {
        const srcDirectory = path.join(__dirname, '..', 'helper');
        const name = this.configuration.output.type === 'script' ? 'default' : 'noop';
        return path.join(srcDirectory, `${name}${this.configuration.ext}`);
    }
    async start() {
        await this.outputHelperScript();
        await this.startWatcher();
    }
    async stop() {
        await this.stopWatcher();
    }
    async outputHelperScript() {
        if (this.configuration.output.type === 'script') {
            await copyFile(this.helperPath, this.configuration.output.path);
        }
    }
    async processCSS(filePath) {
        if (0 < this.tasks.size) {
            await this.waitCurrentTasks();
        }
        const exposedPromise = this.createExposedPromise();
        const { configuration } = this;
        const postcssResult = await parseCSS_1.parseCSS({
            plugins: configuration.postcssPlugins,
            options: configuration.postcssOptions,
            file: filePath,
        })
            .catch((error) => {
            exposedPromise.resolve();
            throw error;
        });
        const dest = path.join(`${filePath}${configuration.ext}`);
        this.processedFiles.add(filePath);
        const { output } = configuration;
        const code = generateScript_1.generateScript({
            output: dest,
            helper: output.type === 'css' ? this.helperPath : output.path,
            result: extractPluginResult_1.extractPluginResult(postcssResult),
            root: postcssResult.root,
            cssKey: this.configuration.cssKey,
        });
        exposedPromise.resolve();
        return { dest, code };
    }
    async minifyScripts() {
        const files = [...this.processedFiles].map((file) => `${file}${this.configuration.ext}`);
        const { cssKey, output } = this.configuration;
        if (output.type === 'css') {
            await minifyScriptsForCSS_1.minifyScriptsForCSS({ files, cssKey, dest: output.path });
        }
        else {
            await minifyScripts_1.minifyScripts({ files, cssKey, dest: output.path });
        }
    }
    async waitCurrentTasks() {
        await Promise.all([...this.tasks]);
    }
    createExposedPromise() {
        const exposedPromise = createExposedPromise_1.createExposedPromise();
        this.tasks.add(exposedPromise.promise);
        const removeTask = () => this.tasks.delete(exposedPromise.promise);
        exposedPromise.promise.then(removeTask).catch(removeTask);
        return exposedPromise;
    }
    async startWatcher() {
        await this.stopWatcher();
        this.initialTask = [];
        this.log(`watching: ${this.configuration.path.join(', ')}`);
        const onError = this.onError.bind(this);
        this.watcher = chokidar.watch(this.configuration.path, this.configuration.chokidar)
            .on('error', onError)
            .on('add', (file, stats) => {
            this.log(`[add] ${file}`);
            const promise = this.onAdd(file, stats);
            if (this.initialTask) {
                this.initialTask.push(promise);
            }
            promise.catch(onError);
        })
            .on('change', (file, stats) => {
            this.log(`[change] ${file}`);
            this.onChange(file, stats).catch(onError);
        })
            .on('unlink', (file) => {
            this.log(`[unlink] ${file}`);
            this.onUnlink(file).catch(onError);
        });
        await waitForInitialScanCompletion_1.waitForInitialScanCompletion(this.watcher);
        await Promise.all(this.initialTask);
        this.initialTask = null;
        if (!this.configuration.watch) {
            await this.minifyScripts();
            await this.stop();
        }
    }
    log(...messages) {
        write_1.write(this.configuration.stdout, messages);
    }
    logError(...messages) {
        write_1.write(this.configuration.stderr, messages);
    }
    async stopWatcher() {
        if (this.watcher) {
            await this.watcher.close();
            delete this.watcher;
        }
        await Promise.resolve(null);
    }
    onError(error) {
        this.logError(error);
    }
    async onAdd(file, stats) {
        await this.onChange(file, stats);
    }
    async onChange(file, stats) {
        if (!stats) {
            throw new Error(`no stats is given for ${file}.`);
        }
        if (!stats.isFile()) {
            throw new Error(`${file} is not a file.`);
        }
        const { dest, code } = await this.processCSS(file);
        await writeFilep_1.writeFilep(dest, code);
        this.log(`written: ${dest}`);
    }
    async onUnlink(file) {
        const outputPath = path.join(`${file}${this.configuration.ext}`);
        this.processedFiles.delete(file);
        await deleteFile_1.deleteFile(outputPath);
        this.log(`deleted: ${outputPath}`);
    }
}
exports.Session = Session;
